using System;
using System.Collections;
using System.Data;
using System.Data.Common;
using System.Text;
using Vevo.Shared.DataAccess.Implementation;
using Vevo.Shared.Domain.SystemServices;
using Vevo.Shared.SystemServices;
using Vevo.Shared.Utilities;

namespace Vevo.Shared.DataAccess
{
    /// <summary>
    /// Summary description for DataAccess
    /// </summary>
    public static class DataAccess
    {
        public delegate void LogErrorCallback( Exception ex );


        #region Private

        private static IPerRequestDataStorage _transactionStorage = new HttpContextPerRequestDataStorage();

        private static DbTransaction Transaction
        {
            get
            {
                if (_transactionStorage.Items.Contains( "DbTransaction" ))
                    return (DbTransaction) _transactionStorage.Items["DbTransaction"];
                else
                    return null;
            }
            set
            {
                _transactionStorage.Items["DbTransaction"] = value;
            }
        }

        private static DbProviderFactory DbFactory
        {
            get { return DataAccessHelper.GetProviderFactory(); }
        }

        //--------------------------------------------------------------------
        // Private methods
        //--------------------------------------------------------------------

        private static DbCommand CreateCommand()
        {
            DbConnection conn;
            if (Transaction == null)
                conn = DataAccessHelper.CreateConnection();
            else
                conn = Transaction.Connection;

            return conn.CreateCommand();
        }

        private static DbCommand[] SplitBatchCommandNoParameter( DbCommand command )
        {
            string mainText = command.CommandText;

            ArrayList commands = new ArrayList();
            string[] textList = mainText.Split( new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries );

            for (int i = 0; i < textList.Length; i++)
            {
                DbCommand newCommand = command.Connection.CreateCommand();
                newCommand.Connection = command.Connection;
                newCommand.Transaction = command.Transaction;
                newCommand.CommandType = command.CommandType;
                newCommand.CommandText = textList[i] + ";";
                commands.Add( newCommand );
            }

            DbCommand[] results = new DbCommand[commands.Count];
            commands.CopyTo( results );

            return results;
        }

        private static DbCommand[] SplitBatchCommand(
            DbCommand command,
            DbParameter[] parameters )
        {
            string mainText = command.CommandText;

            ArrayList commands = new ArrayList();
            string[] textList = mainText.Split( ';' );

            int currentParameter = 0;
            for (int i = 0; i < textList.Length; i++)
            {
                if (textList[i].Trim() != "")
                {
                    DbCommand newCommand = command.Connection.CreateCommand();
                    newCommand.Connection = command.Connection;
                    newCommand.Transaction = command.Transaction;
                    newCommand.CommandType = command.CommandType;
                    newCommand.CommandText = textList[i] + ";";

                    int numParams = GetNumberOfParameters( textList[i] );

                    string[] nameParameters = null;
                    if (numParams != 0)
                        nameParameters = GetParameterNames( textList[i].Trim() );

                    for (int j = 0;
                         j < numParams;
                         j++)
                    {
                        parameters[currentParameter].ParameterName = nameParameters[j];
                        newCommand.Parameters.Add( parameters[currentParameter] );
                        currentParameter++;
                    }

                    commands.Add( newCommand );
                }
            }
            DbCommand[] results = new DbCommand[commands.Count];
            commands.CopyTo( results );

            return results;
        }

        private static string[] GetParameterNames( string sql )
        {
            ArrayList parameters = new ArrayList();

            TokenProcessor processor = new TokenProcessor( sql );

            processor.SkipTo( '@' );
            while (!processor.IsEndOfString())
            {
                processor.SeekRelative( 1 );
                if (processor.GetRelativeChar( 0 ) != '@')
                {
                    parameters.Add(
                        processor.ExtractTo( TokenProcessor.CharType.NonLetterDigit ) );
                }
                processor.SkipTo( '@' );
            }

            string[] result = new string[parameters.Count];
            parameters.CopyTo( result );
            return result;
        }

        private static int GetNumberOfParameters( string sql )
        {
            int numSymbol = StringUtilities.CountOccurrencesInString( sql, '@' );
            int numNonVar = StringUtilities.CountOccurrencesInString( sql, "@@" );
            return numSymbol - (2 * numNonVar);
        }

        // ExecuteSelect-Related Functions
        public static DataTable ExecuteSelectImplementation(
            DbCommand command,
            params DbParameter[] parameters )
        {
            DataTable table;
            try
            {
                // Jet data provider only supports one SQL command at a time.
                // Need to separate execution for Jet.
                DbCommand[] commandList = SplitBatchCommand( command, parameters );

                DbDataReader reader = null;
                foreach (DbCommand current in commandList)
                {
                    if (reader != null)
                        reader.Close();

                    current.CommandTimeout = 300;
                    reader = current.ExecuteReader();
                }

                // Only return result from the last query
                table = new DataTable();
                table.Load( reader );
                reader.Close();
            }
            catch (Exception ex)
            {
                if (DataAccessHelper.LogError != null)
                    DataAccessHelper.LogError( ex );
                throw ex;
            }
            return table;
        }

        public static DataTable ExecuteSelectNoTransaction(
            DbCommand command,
            params DbParameter[] parameters )
        {
            DataTable table = null;
            try
            {
                command.Connection.Open();
                table = ExecuteSelectImplementation( command, parameters );
            }
            finally
            {
                command.Connection.Close();
            }
            return table;
        }

        public static DataTable ExecuteSelectWithTransaction(
            DbCommand command,
            params DbParameter[] parameters )
        {
            command.Transaction = Transaction;
            return ExecuteSelectImplementation( command, parameters );
        }

        // ExecuteNonQuery-Related Functions
        private static int ExecuteNonQueryImplementation(
            DbCommand command,
            params DbParameter[] parameters )
        {
            int affectedRows = -1;
            try
            {
                // Jet data provider only supports one SQL command at a time.
                // Need to separate execution for Jet.
                DbCommand[] commandList = SplitBatchCommand( command, parameters );

                foreach (DbCommand current in commandList)
                {
                    current.CommandTimeout = 300;
                    affectedRows = current.ExecuteNonQuery();
                }
            }
            catch (Exception ex)
            {
                DataAccessHelper.LogError( ex );
                throw ex;
            }
            return affectedRows;
        }

        private static int ExecuteNonQueryNoTransaction(
            DbCommand command,
            params DbParameter[] parameters )
        {
            int affectedRows = -1;
            try
            {
                command.Connection.Open();
                affectedRows = ExecuteNonQueryImplementation( command, parameters );
            }
            finally
            {
                command.Connection.Close();
            }

            return affectedRows;
        }

        private static int ExecuteNonQueryWithTransaction(
            DbCommand command,
            params DbParameter[] parameters )
        {
            command.Transaction = Transaction;
            return ExecuteNonQueryImplementation( command, parameters );
        }

        // ExecuteScalar-Related Functions
        private static string ExecuteScalarImplementation(
            DbCommand command,
            params DbParameter[] parameters )
        {
            string value = String.Empty;
            try
            {
                // Jet data provider only supports one SQL command at a time.
                // Need to separate execution for Jet.
                DbCommand[] commandList = SplitBatchCommand( command, parameters );

                object result = null;
                foreach (DbCommand current in commandList)
                {
                    current.CommandTimeout = 300;
                    result = current.ExecuteScalar();
                }

                if (result == null)
                    value = String.Empty;
                else
                    value = result.ToString();
            }
            catch (Exception ex)
            {
                DataAccessHelper.LogError( ex );
                throw ex;
            }
            return value;
        }

        private static string ExecuteScalarNoTransaction(
            DbCommand command,
            params DbParameter[] parameters )
        {
            string value = String.Empty;
            try
            {
                command.Connection.Open();
                value = ExecuteScalarImplementation( command, parameters );
            }
            finally
            {
                command.Connection.Close();
            }
            return value;
        }

        private static string ExecuteScalarWithTransaction(
            DbCommand command,
            params DbParameter[] parameters )
        {
            command.Transaction = Transaction;
            return ExecuteScalarImplementation( command, parameters );
        }

        // ExecuteNonQueryNoParameter-Related Functions
        private static int ExecuteNonQueryNoParameterImplementation( DbCommand command )
        {
            int affectedRows = -1;
            string commandTextLine = String.Empty;
            try
            {
                DbCommand[] commandList = SplitBatchCommandNoParameter( command );

                foreach (DbCommand current in commandList)
                {
                    commandTextLine = current.CommandText;
                    current.CommandTimeout = 300;
                    affectedRows = current.ExecuteNonQuery();
                }
            }
            catch (Exception ex)
            {
                DataAccessHelper.LogError( ex );
                throw new Exception( ex.Message + "<br/> Command :" + commandTextLine );
            }
            return affectedRows;
        }

        private static int ExecuteNonQueryNoParameterNoTransaction( DbCommand command )
        {
            int affectedRows = -1;
            try
            {
                command.Connection.Open();
                affectedRows = ExecuteNonQueryNoParameterImplementation( command );
            }
            finally
            {
                command.Connection.Close();
            }

            return affectedRows;
        }

        private static int ExecuteNonQueryNoParameterWithTransaction( DbCommand command )
        {
            command.Transaction = Transaction;
            return ExecuteNonQueryNoParameterImplementation( command );
        }

        private static Exception CreateVerboseException(
            Exception ex,
            DbCommand command,
            DbParameter[] parameters )
        {
            StringBuilder sb = new StringBuilder();

            sb.AppendLine( ex.Message );
            sb.AppendLine( "SQL: " );
            sb.AppendLine( command.CommandText );

            if (parameters != null)
            {
                sb.AppendLine( "Parameters: " );

                for (int i = 0; i < parameters.Length; i++)
                {
                    sb.Append( "\t" );
                    sb.Append( parameters[i].ParameterName );
                    sb.Append( "\t" );
                    sb.Append( parameters[i].Value.ToString() );
                    sb.Append( "\t" );
                    sb.Append( parameters[i].DbType );
                    sb.AppendLine();
                }
            }
            else
            {
                sb.AppendLine();
            }

            return new DataAccessException( sb.ToString(), ex );
        }

        private static DateTime AdjustYearForAccess( DateTime value )
        {
            if (DataAccessHelper.IsMicrosoftAccess())
            {
                // Adjust non-AD calendar for MS Access. 
                // This might be an Access bug. If using non-AD local, e.g. Buddhist calendar, 
                // MS Access will not record the actual year on DateTime object.
                // This is kind of a hack, but there is no clean solution yet.

                int localYear = Convert.ToInt32( value.ToString( "yyyy" ) );
                DateTime dateTime = value.AddYears( localYear - value.Year );

                // MS Access cannot use milliseconds, so need to create a new DateTime object 
                // without milliseconds.

                return new DateTime(
                    dateTime.Year, dateTime.Month, dateTime.Day,
                    dateTime.Hour, dateTime.Minute, dateTime.Second );
            }
            else
            {
                return value;
            }
        }
        #endregion


        //--------------------------------------------------------------------
        // Public methods
        //--------------------------------------------------------------------

        static DataAccess()
        {

        }


        public static DbCommand CreateStoredProcedureCommand()
        {
            DbCommand comm = CreateCommand();
            comm.CommandType = CommandType.StoredProcedure;
            return comm;
        }

        public static DbCommand CreateTextCommand()
        {
            return CreateCommand();
        }

        public static DbParameter CreateParameterBool( bool value )
        {
            DbParameter param = DbFactory.CreateParameter();
            param.DbType = DbType.Boolean;
            param.Value = value;
            return param;
        }

        public static DbParameter CreateParameterInt32( int value )
        {
            DbParameter param = DbFactory.CreateParameter();
            param.DbType = DbType.Int32;
            param.Value = value;
            return param;
        }

        public static DbParameter CreateParameterDecimal( decimal value )
        {
            DbParameter param = DbFactory.CreateParameter();
            param.DbType = DbType.Decimal;
            param.Value = value;
            return param;
        }

        public static DbParameter CreateParameterDecimalMoney( decimal value )
        {
            DbParameter param = DbFactory.CreateParameter();
            param.DbType = DbType.Currency;
            param.Value = Decimal.Round( value, DataAccessHelper.DecimalPlaceNumber, MidpointRounding.AwayFromZero );
            return param;
        }

        public static DbParameter CreateParameterDouble( double value )
        {
            DbParameter param = DbFactory.CreateParameter();
            param.DbType = DbType.Double;
            param.Value = value;
            return param;
        }

        public static DbParameter CreateParameterString( string value )
        {
            DbParameter param = DbFactory.CreateParameter();
            param.DbType = DbType.String;
            param.Value = value;
            return param;
        }

        public static DbParameter CreateParameterString( string value, int size )
        {
            DbParameter param = CreateParameterString( value );
            param.Size = size;
            return param;
        }

        public static DbParameter CreateParameterDateTime( DateTime value )
        {
            DateTime dateTime = AdjustYearForAccess( value );

            DbParameter param = DbFactory.CreateParameter();
            param.DbType = DbType.DateTime;
            param.Value = dateTime;
            return param;
        }

        public static string CreateLiteralBool( bool value )
        {
            return DbImplementation.Sql.CreateBoolText( value );
        }

        public static DataTable ExecuteSelect( string sqlString )
        {
            DbCommand command = CreateTextCommand();
            command.CommandText = sqlString;
            return ExecuteSelect( command );
        }

        public static DataTable ExecuteSelect(
            string sqlString,
            params DbParameter[] parameters )
        {
            DbCommand command = CreateTextCommand();
            command.CommandText = sqlString;
            return ExecuteSelect( command, parameters );
        }

        public static DataTable ExecuteSelect( DbCommand command )
        {
            return ExecuteSelect( command, null );
        }

        public static DataTable ExecuteSelect(
            DbCommand command,
            params DbParameter[] parameters )
        {
            try
            {
                if (Transaction == null)
                    return ExecuteSelectNoTransaction( command, parameters );
                else
                    return ExecuteSelectWithTransaction( command, parameters );
            }
            catch (Exception ex)
            {
                if (DataAccessHelper.ShowVerboseError)
                    throw CreateVerboseException( ex, command, parameters );
                else
                    throw ex;
            }
        }

        public static int ExecuteNonQuery( string sqlString )
        {
            DbCommand command = CreateTextCommand();
            command.CommandText = sqlString;
            return ExecuteNonQuery( command );
        }

        public static int ExecuteNonQuery(
            string sqlString,
            params DbParameter[] parameters )
        {
            DbCommand command = CreateTextCommand();
            command.CommandText = sqlString;
            return ExecuteNonQuery( command, parameters );
        }

        public static int ExecuteNonQuery( DbCommand command )
        {
            return ExecuteNonQuery( command, null );
        }

        public static int ExecuteNonQuery(
            DbCommand command,
            params DbParameter[] parameters )
        {
            try
            {
                if (Transaction == null)
                    return ExecuteNonQueryNoTransaction( command, parameters );
                else
                    return ExecuteNonQueryWithTransaction( command, parameters );
            }
            catch (Exception ex)
            {
                if (DataAccessHelper.ShowVerboseError)
                    throw CreateVerboseException( ex, command, parameters );
                else
                    throw ex;
            }
        }

        public static string ExecuteScalar( string sqlString )
        {
            DbCommand command = CreateTextCommand();
            command.CommandText = sqlString;
            return ExecuteScalar( command );
        }

        public static string ExecuteScalar(
            string sqlString,
            params DbParameter[] parameters )
        {
            DbCommand command = CreateTextCommand();
            command.CommandText = sqlString;
            return ExecuteScalar( command, parameters );
        }

        public static string ExecuteScalar( DbCommand command )
        {
            return ExecuteScalar( command, null );
        }

        public static string ExecuteScalar(
            DbCommand command,
            params DbParameter[] parameters )
        {
            try
            {
                if (Transaction == null)
                    return ExecuteScalarNoTransaction( command, parameters );
                else
                    return ExecuteScalarWithTransaction( command, parameters );
            }
            catch (Exception ex)
            {
                if (DataAccessHelper.ShowVerboseError)
                    throw CreateVerboseException( ex, command, parameters );
                else
                    throw ex;
            }
        }

        public static void BeginTransaction( DbTransaction transaction )
        {
            Transaction = transaction;
        }

        public static void EndTransaction()
        {
            Transaction = null;
        }

        public static DataTable CopyTableRows(
            DataTable source, int startIndex, int endIndex )
        {
            DataTable result = source.Clone();
            for (int i = startIndex; i <= endIndex && i < source.Rows.Count; i++)
            {
                result.ImportRow( source.Rows[i] );
            }
            return result;
        }

        public static int ExecuteNonQueryNoParameter( string sqlString )
        {
            DbCommand command = CreateTextCommand();
            command.CommandText = sqlString;
            return ExecuteNonQueryNoParameter( command );
        }

        public static int ExecuteNonQueryNoParameter( DbCommand command )
        {
            if (Transaction == null)
                return ExecuteNonQueryNoParameterNoTransaction( command );
            else
                return ExecuteNonQueryNoParameterWithTransaction( command );
        }

        public static DbParameter[] ConvertArrayListToDbParameters( ArrayList array )
        {
            DbParameter[] newArray = new DbParameter[array.Count];
            array.CopyTo( newArray );

            return newArray;
        }

        public static bool IsColumnExisted( string tableName, string columnName )
        {
            DataTable table = DataAccess.ExecuteSelect(
                "SELECT * FROM " + tableName + " WHERE 1 = 0;" );

            return table.Columns.Contains( columnName );
        }

        public static bool IsTableExisted( string tableName )
        {
            string result = DataAccess.ExecuteScalar(
                "IF object_id( '" + tableName + "', 'U') IS NOT NULL SELECT 'true' ELSE SELECT 'false'" );

            return bool.Parse( result );
        }

        // Set data storage during startup. This storage will keep sensitive 
        // data that should be unique for each HTTP request/response cycle only, 
        // not accessible globally across the HTTP requests.
        public static void SetUpPerRequestDataStorage( IPerRequestDataStorage Storage )
        {
            _transactionStorage = Storage;
        }
    }
}
