using System;
using System.Collections.Generic;
using System.Data;
using System.Text;

using Vevo.Domain;
using Vevo.Domain.DataInterfaces;
using Vevo.Domain.Orders;
using Vevo.Domain.Products;
using Vevo.Shared.DataAccess;
using Vevo.Shared.Utilities;
using Vevo.Domain.Stores;

namespace Vevo.Data
{
    public class CartItemRepositoryNoCache : ICartItemRepository
    {
        #region Private

        private const string AllCartItemFields =
            " CartItemID, CartID, ProductID, OptionItemIDs, Quantity, " +
            " Recipient, PersonalNote, NeedPhysical, GiftValue ";


        private OptionItemValueCollection CreateOptions( DataRow row )
        {
            return new OptionItemValueCollection(
                Culture.Null, row["OptionItemIDs"].ToString(), row["ProductID"].ToString() );
        }

        private CartItemGiftDetails CreateCartItemGiftDetails( DataRow row )
        {
            return new CartItemGiftDetails(
                row["Recipient"].ToString(),
                row["PersonalNote"].ToString(),
                ConvertUtilities.ToBoolean( row["NeedPhysical"] ),
                ConvertUtilities.ToDecimal( row["GiftValue"] ) );
        }

        private ICartItem CreateOneFromRow( DataRow row )
        {
            Product product = DataAccessContext.ProductRepository.GetOne(
                Culture.Null, row["ProductID"].ToString(),Store.Null.StoreID );

            return new CartItem(
                row["CartItemID"].ToString(),
                row["CartID"].ToString(),
                product,
                ConvertUtilities.ToInt32( row["Quantity"] ),
                CreateOptions( row ),
                CreateCartItemGiftDetails( row )
                );
        }

        private IList<ICartItem> ConvertTableToList( DataTable table)
        {
            return ConvertTableToList( table, 0, table.Rows.Count);
        }

        private IList<ICartItem> ConvertTableToList( DataTable table, int startIndex, int endIndex )
        {
            IList<ICartItem> list = new List<ICartItem>();

            for (int i = startIndex; i <= endIndex && i < table.Rows.Count; i++)
                list.Add( CreateOneFromRow( table.Rows[i] ) );

            return list;
        }


        #endregion


        #region ICartItemRepository Members

        public ICartItem Create( ICartItem cartItem )
        {
            DataAccess.ExecuteNonQuery(
                "INSERT INTO CartItem (CartItemID, CartID, ProductID, OptionItemIDs, Quantity, " +
                "    Recipient, PersonalNote, NeedPhysical, GiftValue) " +
                "VALUES (@CartItemID, @CartID, @ProductID, @OptionItemIDs, @Quantity, " +
                "    @Recipient, @PersonalNote, @NeedPhysical, @GiftValue); ",
                DataAccess.CreateParameterString( cartItem.CartItemID ),
                DataAccess.CreateParameterString( cartItem.CartID ),
                DataAccess.CreateParameterString( cartItem.ProductID ),
                DataAccess.CreateParameterString( cartItem.Options.ToString() ),
                DataAccess.CreateParameterInt32( cartItem.Quantity ),
                DataAccess.CreateParameterString( cartItem.GiftDetails.Recipient ),
                DataAccess.CreateParameterString( cartItem.GiftDetails.PersonalNote ),
                DataAccess.CreateParameterBool( cartItem.GiftDetails.NeedPhysical ),
                DataAccess.CreateParameterDecimal( cartItem.GiftDetails.GiftValue ) );

            return cartItem;
        }

        public ICartItem Update( ICartItem cartItem )
        {
            DataAccess.ExecuteNonQuery(
                "UPDATE CartItem " +
                "SET CartID = @CartID, ProductID = @ProductID, " +
                "    OptionItemIDs = @OptionItemIDs, Quantity = @Quantity, " +
                "    Recipient = @Recipient, PersonalNote = @PersonalNote, " +
                "    NeedPhysical = @NeedPhysical, GiftValue = @GiftValue " +
                "WHERE CartItemID = @CartItemID; ",
                DataAccess.CreateParameterString( cartItem.CartID ),
                DataAccess.CreateParameterString( cartItem.ProductID ),
                DataAccess.CreateParameterString( cartItem.Options.ToString() ),
                DataAccess.CreateParameterInt32( cartItem.Quantity ),
                DataAccess.CreateParameterString( cartItem.GiftDetails.Recipient ),
                DataAccess.CreateParameterString( cartItem.GiftDetails.PersonalNote ),
                DataAccess.CreateParameterBool( cartItem.GiftDetails.NeedPhysical ),
                DataAccess.CreateParameterDecimal( cartItem.GiftDetails.GiftValue ),
                DataAccess.CreateParameterString( cartItem.CartItemID ) );

            return cartItem;
        }

        public void Delete( string cartItemID )
        {
            DataAccess.ExecuteNonQuery(
                "DELETE FROM CartItem " +
                "WHERE CartItemID = @CartItemID; ",
                DataAccess.CreateParameterString( cartItemID )
            );
        }

        public void DeleteByCartID( string cartID )
        {
            DataAccess.ExecuteNonQuery(
                "DELETE FROM CartItem " +
                "WHERE CartID = @CartID; ",
                DataAccess.CreateParameterString( cartID )
            );
        }

        public ICartItem GetOne( string cartItemID)
        {
            DataTable table = DataAccess.ExecuteSelect(
                "SELECT " + AllCartItemFields +
                "FROM CartItem " +
                "WHERE CartItemID = @CartItemID; ",
                DataAccess.CreateParameterString( cartItemID ) );

            if (table.Rows.Count > 0)
                return CreateOneFromRow( table.Rows[0] );
            else
                return NullCartItem.Instance;
        }

        public IList<ICartItem> GetAllByCartID( string cartID)
        {
            DataTable table = DataAccess.ExecuteSelect(
                "SELECT " + AllCartItemFields +
                "FROM CartItem " +
                "WHERE CartID = @CartID; ",
                DataAccess.CreateParameterString( cartID )
            );

            return ConvertTableToList( table);
        }

        #endregion
    }
}
