using System;
using System.Data;
using System.Collections.Generic;
using System.Configuration;
using System.Text;

namespace Vevo.WebUI.Ajax
{
    public class AjaxPostbackBlock
    {
        private string _originalText;
        private int _startIndex;
        private int _currentIndex;

        private int _bodyLength;
        private string _type;
        private string _id;
        private string _body;


        private int GetNextPipeIndex( int startIndex )
        {
            int index = _originalText.IndexOf( '|', startIndex );
            if (index == -1)
                throw new ApplicationException( "Invalid data format. Cannot find '|' symbol." );

            return index;
        }

        private string ExtractHeader()
        {
            return _originalText.Substring( _currentIndex, GetNextPipeIndex( _currentIndex ) - _currentIndex );
        }

        private string EatHeader()
        {
            _currentIndex = GetNextPipeIndex( _currentIndex ) + 1;
            return ExtractHeader();
        }

        private string ExtractBody()
        {
            return _originalText.Substring( _currentIndex, _bodyLength );
        }

        private void EatBody()
        {
            _currentIndex += _bodyLength + 1;       // Go pass the pipe character
        }

        private Exception CreateInvalidBodyLengthException()
        {
            string message = String.Format(
                "Invalid body lenghth for Ajax postback data. Expected = {0}, Actual = {1}",
                _bodyLength, _body.Length );

            return new ApplicationException( message );
        }

        public int BodyLength
        {
            get { return _bodyLength; }
        }

        public string Type
        {
            get { return _type; }
            set { _type = value; }
        }

        public string ID
        {
            get { return _id; }
            set { _id = value; }
        }

        public string Body
        {
            get { return _body; }
            set
            {
                _body = value;
                _bodyLength = _body.Length;
            }
        }

        public int TextBlockLength
        {
            get { return _currentIndex - _startIndex; }
        }

        public AjaxPostbackBlock()
        {
        }

        public AjaxPostbackBlock( string text, int startIndex )
        {
            _originalText = text;
            _startIndex = startIndex;
            _currentIndex = startIndex;

            ParseText();
        }

        public void ParseText()
        {
            _bodyLength = Convert.ToInt32( ExtractHeader() );
            EatHeader();

            _type = ExtractHeader();
            EatHeader();

            _id = ExtractHeader();
            EatHeader();

            _body = ExtractBody();
            EatBody();

            if (_bodyLength != _body.Length)
                throw CreateInvalidBodyLengthException();
        }

        public void AppendTo( StringBuilder sb )
        {
            sb.Append( BodyLength );
            sb.Append( '|' );
            sb.Append( Type );
            sb.Append( '|' );
            sb.Append( ID );
            sb.Append( '|' );
            sb.Append( Body );
            sb.Append( '|' );
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            AppendTo( sb );
            return sb.ToString();
        }

    }

    /// <summary>
    /// Summary description for AjaxPostbackData
    /// </summary>
    public class AjaxPostbackData
    {
        private string _originalText;
        private List<AjaxPostbackBlock> _blocks;


        private void SplitAjaxUpdateContent( string text, out string header, out string body, out string other )
        {
            header = String.Empty;
            body = String.Empty;
            other = String.Empty;

            int headerEnd = FindHeaderEndIndex( text );

            header = text.Substring( 0, headerEnd );

            int length = Convert.ToInt32( header.Split( '|' )[0] );

            body = text.Substring( headerEnd, length );
            other = text.Substring( headerEnd + length );
        }

        private static int FindHeaderEndIndex( string text )
        {
            int current = 0;
            for (int i = 0; i < 3; i++)
            {
                current = text.IndexOf( '|', current );
                if (current == -1)
                    return -1;

                current++;
            }

            return current;
        }

        private void ParseTextData()
        {
            _blocks = new List<AjaxPostbackBlock>();

            int currentIndex = 0;
            while (currentIndex < _originalText.Length)
            {
                AjaxPostbackBlock block = new AjaxPostbackBlock( _originalText, currentIndex );
                _blocks.Add( block );

                currentIndex += block.TextBlockLength;
            }
        }


        public List<AjaxPostbackBlock> Blocks
        {
            get
            {
                return _blocks;
            }
        }


        public AjaxPostbackData( string ajaxPostback )
        {
            _originalText = ajaxPostback;

            ParseTextData();
        }

        public static bool TryParse( string text, out AjaxPostbackData ajaxData )
        {
            try
            {
                ajaxData = new AjaxPostbackData( text );
                return true;
            }
            catch
            {
                ajaxData = null;
                return false;
            }
        }

        public static bool IsValidData( string text )
        {
            return FindHeaderEndIndex( text ) >= 0;
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();

            foreach (AjaxPostbackBlock block in Blocks)
            {
                block.AppendTo( sb );
            }

            return sb.ToString();
        }

    }

}
